import os
import subprocess
import sys

import pytest
from click.testing import CliRunner

from vja.cli import cli


@pytest.fixture(name='runner', scope='session')
def setup_runner():
    return CliRunner()


def invoke(runner, command, return_code=0, user_input=None, catch_exceptions=True):
    if isinstance(command, str):
        command = command.split()
    res = runner.invoke(cli, command, input=user_input, catch_exceptions=catch_exceptions)
    assert res.exit_code == return_code, res
    return res


def _login_as_test_user():
    result = subprocess.run('vja logout'.split(), check=False)
    if result.returncode:
        print(result.stdout)
        print(result.stderr)
        return result.returncode

    result = subprocess.run('vja --username=test --password=test user show'.split(), check=False)
    if result.returncode:
        print(result.stdout)
        print(result.stderr)
    return result.returncode


def _create_list_and_task():
    result = subprocess.run('vja list add test-list'.split(), check=False)
    if result.returncode:
        print(result.stdout)
        print(result.stderr)
        return result.returncode
    result = subprocess.run('vja add At least one task --force-create '
                            '--priority=5 --due-date=today --tag=my_tag --favorite=True'.split(), check=False)
    if result.returncode:
        print(result.stdout)
        print(result.stderr)
        return result.returncode
    result = subprocess.run('vja add A task without a label --force-create'.split(), check=False)
    if result.returncode:
        print(result.stdout)
        print(result.stderr)
    return result.returncode


def pytest_configure():
    if 'VJA_CONFIGDIR' not in os.environ:
        print('!!! Precondition not met. You must set VJA_CONFIGDIR in environment variables !!!')
        sys.exit(1)

    if _login_as_test_user() > 0:
        print('!!! Precondition not met. Cannot connect to Vikunja with user test/test')
        sys.exit(1)

    if _create_list_and_task() > 0:
        print('!!! Unable to create default list')
        sys.exit(1)



class VjaError(Exception):
    def __init__(self, msg):
        super().__init__()
        self.msg = msg

    def __repr__(self):
        return f'VjaError("{self.msg}")'

    def __str__(self):
        return self.msg



import json
import re

from tests.conftest import invoke


class TestUser:
    def test_user_show(self, runner):
        res = invoke(runner, 'user show')
        assert re.search(r'username=\'test\'', res.output)

    def test_user_show_json(self, runner):
        res = invoke(runner, 'user show --json')
        assert json.loads(res.output)['username'] == 'test'

    def test_user_show_jsonvja(self, runner):
        res = invoke(runner, 'user show --jsonvja')
        assert json.loads(res.output)['username'] == 'test'


class TestNamespace:
    def test_namespace_ls(self, runner):
        res = invoke(runner, 'namespace ls')
        assert re.search(r'test\'s namespace', res.output)

    def test_namespace_ls_json(self, runner):
        res = invoke(runner, 'namespace ls --json')
        assert json.loads(res.output)[1]['title'] == 'test'

    def test_namespace_ls_jsonvja(self, runner):
        res = invoke(runner, 'namespace ls --jsonvja')
        assert json.loads(res.output)[1]['title'] == 'test'

    def test_namespace_ls_custom_format(self, runner):
        res = invoke(runner, 'namespace ls --custom-format=ids_only')
        for line in res.output:
            assert re.match(r'^-?\d*$', line)


class TestList:
    def test_list_ls(self, runner):
        res = invoke(runner, 'list ls')
        assert re.search(r'test-list', res.output)

    def test_list_show(self, runner):
        res = invoke(runner, 'list show 1')
        assert len(res.output) > 0

    def test_list_ls_custom_format(self, runner):
        res = invoke(runner, 'list ls --custom-format=ids_only')
        for line in res.output:
            assert re.match(r'^\d*$', line)


class TestBucket:
    def test_bucket_ls(self, runner):
        res = invoke(runner, 'bucket ls --list-id=1')
        assert re.search(r'Backlog', res.output)

    def test_bucket_ls_custom_format(self, runner):
        res = invoke(runner, 'bucket ls --list-id=1 --custom-format=ids_only')
        for line in res.output:
            assert re.match(r'^\d*$', line)


class TestLabel:
    def test_label_ls(self, runner):
        res = invoke(runner, 'label ls')
        assert re.search(r'my_tag', res.output)

    def test_label_ls_custom_format(self, runner):
        res = invoke(runner, 'label ls --custom-format=ids_only')
        for line in res.output:
            assert re.match(r'^\d*$', line)


class TestSingleTask:
    def test_task_show(self, runner):
        res = invoke(runner, 'show 1')
        assert re.search(r'id: 1', res.output)

    def test_task_show_json(self, runner):
        res = invoke(runner, 'show 1 --json')
        assert json.loads(res.output)['id'] == 1

    def test_task_show_jsonvja(self, runner):
        res = invoke(runner, 'show 1 --jsonvja')
        data = json.loads(res.output)
        assert data['id'] == 1
        assert data['title'] is not None
        assert data['position'] is not None
        assert data['kanban_position'] is not None
        assert data['bucket_id'] is not None
        assert data['tasklist']['id'] is not None
        assert data['created'] is not None
        assert data['updated'] is not None


class TestTaskList:
    def test_task_ls(self, runner):
        res = invoke(runner, 'ls')
        assert re.search(r'At least one task', res.output)

    def test_task_ls_json(self, runner):
        res = invoke(runner, 'ls --json')
        assert json.loads(res.output)[0]['title'] is not None

    def test_task_ls_jsonvja(self, runner):
        res = invoke(runner, 'ls --jsonvja')
        data = json.loads(res.output)[0]
        assert data['title'] is not None

    def test_task_filter_bucket(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--bucket=1'])
        data = json.loads(res.output)
        assert len(data) > 0
        assert all(i['bucket_id'] == 1 for i in data)
        res = invoke(runner, ['ls', '--jsonvja', '--bucket=9999'])
        data = json.loads(res.output)
        assert len(data) == 0

    def test_task_filter_due(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--due-date=after yesterday'])
        data = json.loads(res.output)
        assert len(data) > 0
        res = invoke(runner, ['ls', '--jsonvja', '--due-date=before tomorrow'])
        data = json.loads(res.output)
        assert len(data) > 0
        res = invoke(runner, ['ls', '--jsonvja', '--due-date=after next week'])
        data = json.loads(res.output)
        assert len(data) == 0
        res = invoke(runner, ['ls', '--jsonvja', '--due-date='''''''])
        data = json.loads(res.output)
        assert len(data) > 0
        assert all(i['due_date'] is None for i in data)

    def test_task_filter_favorite(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--favorite=True'])
        data = json.loads(res.output)
        assert len(data) > 0
        assert all(i['is_favorite'] for i in data)

    def test_task_filter_label(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--label=my_tag'])
        data = json.loads(res.output)
        assert len(data) > 0
        assert data[0]['labels'][0]['title'] == 'my_tag'
        res = invoke(runner, ['ls', '--jsonvja', '--label=Not created'])
        data = json.loads(res.output)
        assert len(data) == 0

    def test_task_filter_label_empty(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--label='''''''])
        data = json.loads(res.output)
        assert len(data) > 0
        assert all(len(i['labels']) == 0 for i in data)

    def test_task_filter_list(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--list=test-list'])
        data = json.loads(res.output)
        assert len(data) > 0
        assert all(i['tasklist']['title'] == 'test-list' for i in data)
        res = invoke(runner, ['ls', '--jsonvja', '--list=Not created'])
        data = json.loads(res.output)
        assert len(data) == 0

    def test_task_filter_namespace(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--namespace=test'])
        data = json.loads(res.output)
        assert len(data) > 0
        assert all(i['tasklist']['namespace']['title'] == 'test' for i in data)
        res = invoke(runner, ['ls', '--jsonvja', '--namespace=Not created'])
        data = json.loads(res.output)
        assert len(data) == 0

    def test_task_filter_priority(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--priority=eq 5'])
        data = json.loads(res.output)
        assert len(data) > 0
        assert all(i['priority'] == 5 for i in data)
        res = invoke(runner, ['ls', '--jsonvja', '--priority=gt 4'])
        data = json.loads(res.output)
        assert len(data) > 0
        assert all(i['priority'] > 4 for i in data)
        res = invoke(runner, ['ls', '--jsonvja', '--priority=gt 5'])
        data = json.loads(res.output)
        assert len(data) == 0

    def test_task_filter_title(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--title=At least one'])
        data = json.loads(res.output)
        assert len(data) > 0
        assert all('At least one task' in i['title'] for i in data)
        res = invoke(runner, ['ls', '--jsonvja', '--title=Not created'])
        data = json.loads(res.output)
        assert len(data) == 0

    def test_task_filter_urgency(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--urgency=10'])
        data = json.loads(res.output)
        assert len(data) > 0
        assert all(i['urgency'] >= 10 for i in data)
        res = invoke(runner, ['ls', '--jsonvja', '--urgency=10000'])
        data = json.loads(res.output)
        assert len(data) == 0

    def test_sort_id(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--sort=id'])
        data = json.loads(res.output)
        assert data[0]['id'] == 1
        res = invoke(runner, ['ls', '--jsonvja', '--sort=-id'])
        data = json.loads(res.output)
        assert data[0]['id'] > 1

    def test_sort_combined(self, runner):
        res = invoke(runner, ['ls', '--jsonvja', '--sort=due_date, is_favorite, -priority, tasklist.title'])
        data = json.loads(res.output)
        assert data[0]['due_date'] is not None
        assert data[-1]['due_date'] is None

    def test_task_custom_format(self, runner):
        res = invoke(runner, 'ls --custom-format=ids_only')
        for line in res.output:
            assert re.match(r'^\d*$', line)



import functools
import logging

from vja.apiclient import ApiClient
from vja.filter import create_filters
from vja.list_service import ListService
from vja.model import Namespace, Label, User, Bucket

logger = logging.getLogger(__name__)
DEFAULT_SORT_STRING = 'done, -urgency, due_date, -priority, tasklist.title, title'


def rgetattr(obj, path: str, *default):
    attrs = path.split('.')
    try:
        return functools.reduce(getattr, attrs, obj)
    except AttributeError:
        if default:
            return default[0]
        raise


class QueryService:
    def __init__(self, list_service: ListService, api_client: ApiClient):
        self._list_service = list_service
        self._api_client = api_client

    # user
    def find_current_user(self):
        return User.from_json(self._api_client.get_user())

    # namespace
    def find_all_namespaces(self):
        return Namespace.from_json_array(self._api_client.get_namespaces())

    # list
    def find_all_lists(self):
        return [self._list_service.convert_list_json(list_json) for list_json in (self._api_client.get_lists())]

    def find_list_by_id(self, list_id):
        return self._list_service.convert_list_json(self._api_client.get_list(list_id))

    # bucket
    def find_all_buckets_in_list(self, list_id):
        return Bucket.from_json_array(self._api_client.get_buckets(list_id))

    # label
    def find_all_labels(self):
        return Label.from_json_array(self._api_client.get_labels())

    # tasks
    def find_filtered_tasks(self, include_completed, sort_string, filter_args):
        task_object_array = [self._list_service.task_from_json(x) for x in
                             self._api_client.get_tasks(exclude_completed=not include_completed)]
        filtered_tasks = self._filter(task_object_array, filter_args)
        return self._sort(filtered_tasks, sort_string)

    def find_task_by_id(self, task_id: int):
        return self._list_service.task_from_json(self._api_client.get_task(task_id))

    @staticmethod
    def _filter(task_object_array, filter_args):
        filters = create_filters(filter_args)
        return list(filter(lambda x: all(f(x) for f in filters), task_object_array))

    @staticmethod
    def _sort(filtered_tasks, sort_string):
        sort_string = sort_string or DEFAULT_SORT_STRING
        sort_values = [(x.strip().strip('-'), x.strip().startswith('-')) for x in sort_string.split(',')]
        for sort_value in reversed(sort_values):
            field_name = 'sortable_due_date' if sort_value[0] == 'due_date' else sort_value[0]
            filtered_tasks.sort(key=lambda x, field=field_name: rgetattr(x, field), reverse=sort_value[1])
        return filtered_tasks



import logging
from typing import Optional

from vja import VjaError
from vja.apiclient import ApiClient
from vja.model import Namespace, List, Task, Label, User

logger = logging.getLogger(__name__)


class ListService:
    def __init__(self, api_client: ApiClient):
        self._api_client = api_client
        self._namespace_by_id: Optional[dict] = None
        self._list_by_id: Optional[dict] = None

    def find_namespace_by_id(self, namespace_id: int) -> Namespace:
        if not self._namespace_by_id:
            self._namespace_by_id = {x['id']: Namespace.from_json(x) for x in self._api_client.get_namespaces()}
        namespace_object = self._namespace_by_id.get(namespace_id)
        if not namespace_object:
            logger.warning(
                'Inconsistent data: namespace_id %s is referring to non existing cached Namespace.', str(namespace_id))
        return namespace_object

    def find_list_by_id(self, list_id: int) -> List:
        if not self._list_by_id:
            self._list_by_id = {x['id']: self.convert_list_json(x) for x in self._api_client.get_lists()}
        return self._list_by_id.get(list_id)

    def find_list_by_title(self, title):
        list_objects = [self.convert_list_json(x) for x in self._api_client.get_lists()]
        if not list_objects:
            raise VjaError('No lists exist. Go and create at least one.')
        list_found = [x for x in list_objects if x.title == title]
        if not list_found:
            raise VjaError(f'List with title {title} does not exist.')
        return list_found[0]

    def get_default_list(self) -> List:
        user = User.from_json(self._api_client.get_user())
        list_found = self.find_list_by_id(user.default_list_id)
        if not list_found:
            list_objects = [self.convert_list_json(x) for x in self._api_client.get_lists()]
            if not list_objects:
                raise VjaError('No lists exist. Go and create at least one.')
            list_objects.sort(key=lambda x: x.id)
            favorite_lists = [x for x in list_objects if x.is_favorite]
            if favorite_lists:
                list_found = favorite_lists[0]
            else:
                list_found = list_objects[0]
        return list_found

    def convert_list_json(self, list_json: dict) -> List:
        namespace = self.find_namespace_by_id(list_json['namespace_id'])
        return List.from_json(list_json, namespace)

    def task_from_json(self, task_json: dict) -> Task:
        list_object = self.find_list_by_id(task_json['list_id'])
        labels = Label.from_json_array(task_json['labels'])
        return Task.from_json(task_json, list_object, labels)



import time
from datetime import datetime

import dateutil.parser
from dateutil import tz
from parsedatetime import parsedatetime


def parse_date_arg_to_datetime(text: str):
    if not text:
        return None
    try:
        return dateutil.parser.isoparse(text)
    except ValueError:
        timetuple, pdt_context = parsedatetime.Calendar(version=parsedatetime.VERSION_CONTEXT_STYLE).parse(text)
        datetime_date = datetime.fromtimestamp(time.mktime(timetuple))
        if not pdt_context.hasTime:
            datetime_date = datetime_date.replace(hour=0, minute=0, second=0)
        return datetime_date


def parse_date_arg(text: str):
    result = parse_date_arg_to_datetime(text)
    return result.astimezone(tz.tzlocal()).isoformat() if result else None



import logging
import time

from vja import VjaError
from vja.apiclient import ApiClient
from vja.list_service import ListService
from vja.model import Label
from vja.parse import parse_date_arg

logger = logging.getLogger(__name__)


class CommandService:
    def __init__(self, list_service: ListService, api_client: ApiClient):
        self._list_service = list_service
        self._api_client = api_client

    def login(self, username, password, totp_passcode):
        self._api_client.authenticate(True, username, password, totp_passcode)

    def logout(self):
        self._api_client.logout()
        logger.info('Logged out')

    # list
    def add_list(self, namespace_id, title):
        if not namespace_id:
            namespaces = self._api_client.get_namespaces()
            namespace_id = min(namespace['id'] if namespace['id'] > 0 else 99999 for namespace in namespaces)
        list_json = self._api_client.put_list(namespace_id, title)
        return self._list_service.convert_list_json(list_json)

    # label
    def add_label(self, title):
        label_json = self._api_client.put_label(title)
        return Label.from_json(label_json)

    # tasks
    _arg_to_json = {'title': {'field': 'title', 'mapping': (lambda x: x)},
                    'note': {'field': 'description', 'mapping': (lambda x: x)},
                    'prio': {'field': 'priority', 'mapping': int},
                    'due': {'field': 'due_date', 'mapping': parse_date_arg},
                    'favorite': {'field': 'is_favorite', 'mapping': bool},
                    'completed': {'field': 'done', 'mapping': bool},
                    'position': {'field': 'position', 'mapping': int},
                    'list_id': {'field': 'list_id', 'mapping': int},
                    'bucket_id': {'field': 'bucket_id', 'mapping': int},
                    'kanban_position': {'field': 'kanban_position', 'mapping': int},
                    'reminder': {'field': 'reminder_dates', 'mapping': (lambda x: [parse_date_arg(x)])}
                    }

    def _args_to_payload(self, args: dict):
        payload = {}
        for arg_name, arg_value in args.items():
            mapper = self._arg_to_json[arg_name]
            payload[mapper['field']] = mapper['mapping'](arg_value)
        return payload

    def add_task(self, title, args: dict):
        args.update({'title': title})
        if args.get('list_id'):
            list_arg = args.pop('list_id')
            if str(list_arg).isdigit():
                list_id = list_arg
            else:
                list_id = self._list_service.find_list_by_title(list_arg).id
        else:
            list_id = self._list_service.get_default_list().id
        tag_name = args.pop('tag') if args.get('tag') else None
        is_force = args.pop('force_create') if args.get('force_create') is not None else False
        if args.get('reminder') == 'due':
            args.update({'reminder': args.get('due') or 'tomorrow'})
        payload = self._args_to_payload(args)
        # workaround: api server does not seem to set positions. But they should not be 0, because sorting between 0s
        # does not work.
        seconds = int(time.time() / 60)
        payload.update({'position': seconds, 'kanban_position': seconds})

        if not is_force:
            self._validate_add_task(title, tag_name)
        logger.debug('put task: %s', payload)
        task_json = self._api_client.put_task(list_id, payload)
        task = self._list_service.task_from_json(task_json)

        label = self._label_from_name(tag_name, is_force) if tag_name else None
        if label:
            self._api_client.add_label_to_task(task.id, label.id)
        return task

    def edit_task(self, task_id: int, args: dict):
        task_remote = self._api_client.get_task(task_id)
        tag_name = args.pop('tag') if args.get('tag') else None
        is_force = args.pop('force_create') if args.get('force_create') is not None else False
        if args.get('reminder') == 'due':
            if args.get('due'):
                args.update({'reminder': args.get('due')})
            else:
                args.update({'reminder': task_remote['due_date'] if task_remote['due_date'] else 'tomorrow'})
        if args.get('note_append'):
            append_note = args.pop('note_append')
            args.update({
                'note': task_remote['description'] + '\n' + append_note if task_remote['description'] else append_note
            })
        payload = self._args_to_payload(args)
        logger.debug('post task: %s', payload)

        task_remote.update(payload)

        task_json = self._api_client.post_task(task_id, task_remote)
        task = self._list_service.task_from_json(task_json)

        label = self._label_from_name(tag_name, is_force) if tag_name else None
        if label:
            if task.has_label(label):
                self._api_client.remove_label_from_task(task.id, label.id)
            else:
                self._api_client.add_label_to_task(task.id, label.id)
        return task

    def toggle_task_done(self, task_id):
        task_remote = self._api_client.get_task(task_id)
        task_remote.update({'done': not task_remote['done']})
        task_json = self._api_client.post_task(task_id, task_remote)
        return self._list_service.task_from_json(task_json)

    def _label_from_name(self, name, is_force):
        if not name:
            return None
        labels_remote = Label.from_json_array(self._api_client.get_labels())
        label_found = [label for label in labels_remote if label.title == name]
        if not label_found:
            if is_force:
                return Label.from_json(self._api_client.put_label(name))
            logger.warning("Ignoring non existing label [%s]. You may want to execute \"label add\" first.", name)
            return None
        return label_found[0]

    def _validate_add_task(self, title, tag_name):
        tasks_remote = self._api_client.get_tasks(exclude_completed=True)
        if any(task for task in tasks_remote if task['title'] == title):
            raise VjaError("Task with title does exist. You may want to run with --force-create.")
        if tag_name:
            labels_remote = Label.from_json_array(self._api_client.get_labels())
            if not any(label for label in labels_remote if label.title == tag_name):
                raise VjaError(
                    "Label does not exist. You may want to execute \"label add\" or run with --force-create.")



import datetime
import json
import re

from tests.conftest import invoke
from vja.cli import cli

ADD_SUCCESS_PATTERN = re.compile(r'.*Created task (\d+) in list .*')
TODAY = datetime.datetime.now()
TOMORROW = (datetime.datetime.now() + datetime.timedelta(days=1))
TODAY_ISO = TODAY.isoformat()
TOMORROW_ISO = TOMORROW.isoformat()


class TestAddTask:
    def test_list_id(self, runner):
        res = invoke(runner, 'add title of new task --force --list=1')
        after = json_for_created_task(runner, res.output)
        assert after['tasklist']['id'] == 1

    def test_list_title(self, runner):
        res = invoke(runner, 'add title of new task --force --list=test-list')
        after = json_for_created_task(runner, res.output)
        assert after['tasklist']['title'] == 'test-list'

    def test_duplicate_task_title_rejected(self, runner):
        invoke(runner, 'add title of new task', 1)

    def test_positions_not_null(self, runner):
        res = invoke(runner, 'add any other new task --force')
        after = json_for_created_task(runner, res.output)
        assert after['kanban_position'] > 0
        assert after['position'] > 0

    def test_default_reminder_uses_due(self, runner):
        res = invoke(runner, 'add title of new task --force --list=test-list --due=today --reminder')
        after = json_for_created_task(runner, res.output)
        assert after['due_date'] == after['reminder_dates'][0]

    def test_default_reminder_with_missing_due_uses_tomorrow(self, runner):
        res = invoke(runner, 'add title of new task --force --list=test-list --reminder')
        after = json_for_created_task(runner, res.output)
        assert after['due_date'] is None
        assert TOMORROW_ISO[0:10] in after['reminder_dates'][0]


class TestEditGeneral:
    def test_edit_title(self, runner):
        before = json_for_task_id(runner, 1)
        new_title = f'{before["title"]}42'
        res = runner.invoke(cli, ['edit', '1', '-i', f'{new_title}'])
        assert res.exit_code == 0, res

        after = json_for_task_id(runner, 1)
        assert after['title'] == new_title
        assert after['updated'] > before['updated']
        # other attributes remain in place
        assert after['due_date'] == before['due_date']
        assert after['reminder_dates'] == before['reminder_dates']
        assert after['position'] == before['position']
        assert after['tasklist']['id'] == before['tasklist']['id']
        assert after['created'] == before['created']

    def test_edit_due_date(self, runner):
        before = json_for_task_id(runner, 1)

        invoke(runner, 'edit 1 --due=today')

        after = json_for_task_id(runner, 1)
        assert datetime.date.today().isoformat()[0:10] in after['due_date']
        assert after['updated'] >= before['updated']

    def test_unset_due_date(self, runner):
        before = json_for_task_id(runner, 1)
        assert before['due_date'] is not None

        invoke(runner, 'edit 1 --due-date=')

        after = json_for_task_id(runner, 1)
        assert after['due_date'] is None
        assert after['updated'] >= before['updated']

    def test_toggle_label(self, runner):
        labels_0 = json_for_task_id(runner, 1)['labels']
        invoke(runner, 'edit 1 --tag=tag1 --force-create')
        labels_1 = json_for_task_id(runner, 1)['labels']
        invoke(runner, 'edit 1 --tag=tag1')
        labels_2 = json_for_task_id(runner, 1)['labels']

        assert labels_0 != labels_1
        assert labels_0 == labels_2
        assert self._has_label_with_title(labels_0, 'tag1') or self._has_label_with_title(labels_1, 'tag1')

    def test_append_note(self, runner):
        invoke(runner, 'edit 1 --note=line1')
        note_1 = json_for_task_id(runner, 1)['description']
        invoke(runner, 'edit 1 --note-append=line2')
        note_2 = json_for_task_id(runner, 1)['description']

        assert note_1 == 'line1'
        assert note_2 == 'line1\nline2'

    def test_edit_list(self, runner):
        invoke(runner, 'list add another list')
        invoke(runner, 'edit 1 --list-id=1')
        list_1 = json_for_task_id(runner, 1)['tasklist']['id']
        invoke(runner, 'edit 1 -l 2')
        list_2 = json_for_task_id(runner, 1)['tasklist']['id']

        assert list_1 == 1
        assert list_2 == 2

    @staticmethod
    def _has_label_with_title(labels, title):
        label_titles = [x['title'] for x in labels]
        return title in label_titles


class TestEditReminder:
    def test_set_reminder_to_task(self, runner):
        invoke(runner, f'edit 2 --due-date={TOMORROW_ISO}')
        invoke(runner, f'edit 2 --reminder={TODAY_ISO}')
        before = json_for_task_id(runner, 2)

        invoke(runner, 'edit 2 --reminder')
        after = json_for_task_id(runner, 2)
        assert before['reminder_dates'][0][:10] == TODAY.date().isoformat()
        assert after['reminder_dates'][0][:10] == TOMORROW.date().isoformat()

    def test_unset_reminder(self, runner):
        invoke(runner, f'edit 2 --reminder={TODAY_ISO}')
        before = json_for_task_id(runner, 2)

        invoke(runner, 'edit 2 --reminder=')
        after = json_for_task_id(runner, 2)
        assert before['reminder_dates'][0][:10] == TODAY.date().isoformat()
        assert after['reminder_dates'][0] is None

    def test_set_reminder_to_due(self, runner):
        invoke(runner, f'edit 2 --reminder={TODAY_ISO}')
        before = json_for_task_id(runner, 2)

        invoke(runner, f'edit 2 --due={TOMORROW_ISO} --reminder=due')
        after = json_for_task_id(runner, 2)
        assert before['reminder_dates'][0][:10] == TODAY.date().isoformat()
        assert after['reminder_dates'][0][:10] == TOMORROW.date().isoformat()


class TestToggleDoneTask:
    def test_toggle_done(self, runner):
        done_0 = json_for_task_id(runner, 1)['done']
        invoke(runner, 'check 1')
        done_1 = json_for_task_id(runner, 1)['done']
        invoke(runner, 'check 1')
        done_2 = json_for_task_id(runner, 1)['done']
        assert done_0 != done_1
        assert done_0 == done_2

    def test_toggle_does_not_modify_other_fields(self, runner):
        invoke(runner, 'edit 1 --priority=5')
        json_0 = json_for_task_id(runner, 1)
        invoke(runner, 'check 1')
        json_1 = json_for_task_id(runner, 1)
        invoke(runner, 'check 1')
        json_2 = json_for_task_id(runner, 1)
        assert json_0['priority'] == json_1['priority'] == json_2['priority'] == 5


class TestMultipleTasks:
    def test_edit_three_tasks(self, runner):
        invoke(runner, 'edit 1 2 3 --priority=4')
        for i in range(1, 4):
            after = json_for_task_id(runner, i)
            assert after['priority'] == 4
            assert after['updated'][:10] == TODAY_ISO[:10]

    def test_show_three_tasks(self, runner):
        res = invoke(runner, 'show 1 2 3')
        assert res.output.count('\n') >= 30


def json_for_created_task(runner, message):
    assert re.match(ADD_SUCCESS_PATTERN, message), message
    task_id = ADD_SUCCESS_PATTERN.findall(message)[0]
    return json_for_task_id(runner, task_id)


def json_for_task_id(runner, task_id):
    res = invoke(runner, f'show {task_id} --jsonvja')

    assert res.exit_code == 0, res
    data = json.loads(res.output)
    return data



import json
import logging
import re

import pytest

from tests.conftest import invoke
from vja.config import VjaConfiguration


class TestBasicOptions:
    def test_version(self, runner):
        res = invoke(runner, '--version')
        assert re.match(r'.*version \d+\.\d+\.\d+', res.output), res.output

    def test_no_args_displays_help(self, runner):
        res = invoke(runner, '')
        assert 'Usage: cli [OPTIONS] COMMAND [ARGS]...' in res.output
        assert res.output.count('\n') > 20

    def test_help_ls(self, runner):
        res = invoke(runner, 'ls --help')
        assert 'Usage: cli ls [OPTIONS]' in res.output
        assert 'list tasks' in res.output
        assert res.output.count('\n') > 12

    def test_verbose_logging(self, runner, caplog):
        caplog.set_level(logging.DEBUG)
        invoke(runner, '-v ls')
        assert 'Read config from' in caplog.text
        assert 'Connecting to api_url' in caplog.text


class TestLoginLogout:
    def test_logout_login_given_password(self, runner, caplog):
        invoke(runner, 'logout')
        assert 'Logged out' in caplog.text
        invoke(runner, '-u test -p test user show')

    def test_logout_login_wrong_given_password(self, runner, caplog):
        invoke(runner, 'logout')
        assert 'Logged out' in caplog.text
        with pytest.raises(Exception) as exception:
            invoke(runner, '-u testxx -p testxx user show', return_code=1)
            assert '412 Client Error' in exception.value
        invoke(runner, '-u test -p test user show')

    def test_logout_login_password_from_stdin(self, runner, caplog):
        invoke(runner, 'logout')
        assert 'Logged out' in caplog.text
        invoke(runner, 'user show', user_input='test\ntest\n')
        assert 'Login successful' in caplog.text

    def test_logout_login_wrong_password_from_stdin(self, runner, caplog):
        invoke(runner, 'logout')
        assert 'Logged out' in caplog.text
        with pytest.raises(Exception) as exception:
            invoke(runner, 'user show', user_input='testy\ntesty\n')
            assert '412 Client Error' in exception.value
        invoke(runner, '-u test -p test user show')

    def test_prompt_when_invalid_token(self, runner, caplog):
        self._invalidate_token()
        invoke(runner, 'user show', user_input='test\ntest\n')
        assert 'trying to retrieve new access token' in caplog.text
        assert 'Login successful' in caplog.text

    def test_http_error(self, runner, caplog):
        invoke(runner, 'show 9999', return_code=1)
        assert 'HTTP-Error 404' in caplog.text

    @staticmethod
    def _invalidate_token():
        config = VjaConfiguration()
        with open(config.get_token_file(), 'w', encoding='utf-8') as token_file:
            json.dump({'token': 'slightly outdated...'}, token_file)



import configparser
import logging
import os

from vja import VjaError

logger = logging.getLogger(__name__)

_DEFAULT_CONFIGDIR = os.path.expanduser("~/.vjacli")
_FILENAME = 'vja.rc'
_TOKEN_JSON = 'token.json'


class VjaConfiguration:

    def __init__(self):
        self._directory = os.environ.get('VJA_CONFIGDIR', _DEFAULT_CONFIGDIR)
        self._file = os.path.join(self._directory, _FILENAME)
        self._parser = self._load(self._file)

    @property
    def file(self):
        return self._file

    def get_api_url(self):
        return self._get('application', 'api_url')

    def get_frontend_url(self):
        return self._get('application', 'frontend_url')

    def get_token_file(self):
        return os.path.join(self._directory, _TOKEN_JSON)

    def get_custom_format_string(self, template_key):
        return self._parser.get('output', template_key, fallback=None)

    @staticmethod
    def _load(filepath):
        logger.debug('Read config from %s', os.path.abspath(filepath))
        parser = configparser.RawConfigParser()
        if not parser.read(filepath):
            raise VjaError(f'Could not load config file from {os.path.abspath(filepath)}')
        return parser

    def _get(self, section, option):
        try:
            return self._parser.get(section, option)
        except (configparser.NoSectionError, configparser.NoOptionError) as ex:
            raise VjaError(f'[{section}] [{option}] not specified in {self.file}. Dying.') from ex



import json
import logging

import click

from vja.model import User, List, Task

NAMESPACE_LIST_FORMAT_DEFAULT = '{x.id:5} {x.title:20.20} {x.description:20.20}'

LIST_LIST_FORMAT_DEFAULT = '{x.id:5} {x.title:20.20} {x.description:20.20}  ' \
                           '{x.namespace.title:20.20} {x.namespace.id:5}'

BUCKET_LIST_FORMAT_DEFAULT = '{x.id:5} {x.title:20.20} {x.is_done_bucket:2} {x.limit:3} {x.count_tasks:5}'

LABEL_LIST_FORMAT_DEFAULT = '{x.id:5} {x.title:20.20}'

TASK_LIST_FORMAT_DEFAULT = '{x.id:5} ({x.priority}) {"*" if x.is_favorite else " "} {x.title:50.50} ' \
                           '{x.due_date.strftime("%a %d.%m %H:%M") if x.due_date else "":15.15} ' \
                           '{"R" if x.reminder_dates else " "} {x.tasklist.title:20.20} ' \
                           '{x.label_titles:20.20} {x.urgency:3}'

logger = logging.getLogger(__name__)


class Output:

    def user(self, user: User, is_json, is_jsonvja):
        self._dump(user, is_json, is_jsonvja)

    def list(self, task_list: List, is_json, is_jsonvja):
        self._dump(task_list, is_json, is_jsonvja)

    def task(self, task: Task, is_json, is_jsonvja):
        self._dump(task, is_json, is_jsonvja)

    def namespace_array(self, object_array, is_json, is_jsonvja, custom_format=None):
        line_format = custom_format or NAMESPACE_LIST_FORMAT_DEFAULT
        self._dump_array(object_array, line_format, is_json, is_jsonvja)

    def list_array(self, object_array, is_json, is_jsonvja, custom_format=None):
        line_format = custom_format or LIST_LIST_FORMAT_DEFAULT
        self._dump_array(object_array, line_format, is_json, is_jsonvja)

    def bucket_array(self, object_array, is_json, is_jsonvja, custom_format=None):
        line_format = custom_format or BUCKET_LIST_FORMAT_DEFAULT
        self._dump_array(object_array, line_format, is_json, is_jsonvja)

    def label_array(self, object_array, is_json, is_jsonvja, custom_format=None):
        line_format = custom_format or LABEL_LIST_FORMAT_DEFAULT
        self._dump_array(object_array, line_format, is_json, is_jsonvja)

    def task_array(self, object_array, is_json, is_jsonvja, custom_format=None):
        line_format = custom_format or TASK_LIST_FORMAT_DEFAULT
        self._dump_array(object_array, line_format, is_json, is_jsonvja)

    @staticmethod
    def _dump(element, is_json, is_jsonvja):
        if is_json:
            click.echo(json.dumps(element.json))
        elif is_jsonvja:
            click.echo(json.dumps(element.data_dict(), default=str))
        else:
            click.echo(element)

    @staticmethod
    def _dump_array(object_array, line_format, is_json, is_jsonvja):
        if is_json:
            click.echo(json.dumps([x.json for x in object_array]))
        elif is_jsonvja:
            click.echo(json.dumps([x.data_dict() for x in object_array], default=str))
        else:
            for x in object_array:
                # https://stackoverflow.com/a/53671539/2935741
                # Note: Using eval() is risky, because arbitrary code may be introduced via the configured formatting
                # templates.
                # Do not use custom templates, if you are unsure what you are doing.
                click.echo(eval(f"f'{line_format}'"))



#!/usr/bin/env python

import logging
import webbrowser
from importlib import metadata

import click
from click_aliases import ClickAliasedGroup

from vja.apiclient import ApiClient
from vja.config import VjaConfiguration
from vja.list_service import ListService
from vja.output import Output
from vja.service_command import CommandService
from vja.service_query import QueryService

logger = logging.getLogger(__name__)


class Application:
    def __init__(self):
        self._configuration = VjaConfiguration()
        api_client = ApiClient(self.configuration.get_api_url(), self._configuration.get_token_file())
        list_service = ListService(api_client)
        self._command_service = CommandService(list_service, api_client)
        self._query_service = QueryService(list_service, api_client)
        self._output = Output()

    @property
    def command_service(self) -> CommandService:
        return self._command_service

    @property
    def query_service(self) -> QueryService:
        return self._query_service

    @property
    def output(self) -> Output:
        return self._output

    @property
    def configuration(self) -> VjaConfiguration:
        return self._configuration

    def open_browser(self, task):
        url = self.configuration.get_frontend_url()
        if task:
            url += f'/tasks/{str(task)}'
        webbrowser.open_new_tab(url)


with_application = click.make_pass_decorator(Application, ensure=True)


@click.group(cls=ClickAliasedGroup, context_settings=dict({'help_option_names': ['-h', '--help']}))
@click.pass_context
@click.version_option(metadata.version("vja"))
@click.option('-v', '--verbose', 'verbose', default=False, is_flag=True, help='verbose output')
@click.option('-u', '--username', 'username', help='username for initial login (optional)')
@click.option('-p', '--password', 'password', help='password for initial login (optional)')
@click.option('-t', '--totp-passcode', '--totp_passcode', 'totp_passcode',
              help='time-based one-time passcode from your authenticator app if TOTP is enabled (optional)')
def cli(ctx=None, verbose=None, username=None, password=None, totp_passcode=None):
    if verbose:
        logging.basicConfig(level=logging.DEBUG)
        logger.debug('Verbose mode on')
    else:
        logging.basicConfig(level=logging.INFO)
        logging.getLogger().setLevel(logging.INFO)
    application = Application()
    ctx.obj = application
    if username:
        application.command_service.login(username, password, totp_passcode)


# user
@cli.group('user', help='subcommand: user (see help)')
def user_group():
    pass


@user_group.command('show', help='print current user')
@click.option('is_json', '--json', default=False, is_flag=True, help='print as Vikunja json')
@click.option('is_jsonvja', '--jsonvja', default=False, is_flag=True, help='print as vja application json')
@with_application
def user_show(application, is_json=False, is_jsonvja=False):
    application.output.user(
        application.query_service.find_current_user(), is_json, is_jsonvja)


# namespaces
@cli.group('namespace', help='subcommand: namespace (see help)')
def namespace_group():
    pass


@namespace_group.command('ls', help='print namespaces ... (id; title; description)')
@click.option('is_json', '--json', default=False, is_flag=True,
              help='print as Vikunja json')
@click.option('is_jsonvja', '--jsonvja', default=False, is_flag=True,
              help='print as vja application json')
@click.option('custom_format', '--custom-format',
              help='use formatting string from .vjacli/vja.rc')
@with_application
def namespace_ls(application, is_json, is_jsonvja, custom_format):
    if custom_format:
        custom_format = application.configuration.get_custom_format_string(custom_format)
    application.output.namespace_array(
        application.query_service.find_all_namespaces(), is_json, is_jsonvja, custom_format)


# lists
@cli.group('list', help='subcommand: list (see help)')
def list_group():
    pass


@list_group.command('add', help='add list with title')
@click.option('namespace_id', '-n', '--namespace-id', help='create list in namespace, default: first list found')
@click.argument('title', nargs=-1, required=True)
@with_application
def list_add(application, title, namespace_id=None):
    tasklist = application.command_service.add_list(namespace_id, " ".join(title))
    click.echo(f'Created list {tasklist.id}')


@list_group.command('ls', help='print lists ... (id; title; description; namespace; namespace_id)')
@click.option('is_json', '--json', default=False, is_flag=True,
              help='print as Vikunja json')
@click.option('is_jsonvja', '--jsonvja', default=False, is_flag=True,
              help='print as vja application json')
@click.option('custom_format', '--custom-format',
              help='use formatting string from .vjacli/vja.rc')
@with_application
def list_ls(application, is_json, is_jsonvja, custom_format):
    if custom_format:
        custom_format = application.configuration.get_custom_format_string(custom_format)
    application.output.list_array(
        application.query_service.find_all_lists(), is_json, is_jsonvja, custom_format)


@list_group.command('show', help='show list details')
@click.argument('list_id', required=True, type=click.INT)
@click.option('is_json', '--json', default=False, is_flag=True,
              help='print as Vikunja json')
@click.option('is_jsonvja', '--jsonvja', default=False, is_flag=True,
              help='print as vja application json')
@with_application
def list_show(application, list_id, is_json, is_jsonvja):
    application.output.list(
        application.query_service.find_list_by_id(list_id), is_json, is_jsonvja)


# buckets
@cli.group('bucket', help='subcommand: kanban buckets (see help)')
def bucket_group():
    pass


@bucket_group.command('ls', help='print kanban buckets of given list ... (id; title; is_done; limit; count tasks)')
@click.option('list_id', '-l', '--list', '--list-id', '--list_id', required=True, type=click.INT,
              help='show buckets of list with id')
@click.option('is_json', '--json', default=False, is_flag=True,
              help='print as Vikunja json')
@click.option('is_jsonvja', '--jsonvja', default=False, is_flag=True,
              help='print as vja application json')
@click.option('custom_format', '--custom-format',
              help='use formatting string from .vjacli/vja.rc')
@with_application
def bucket_ls(application, list_id, is_json, is_jsonvja, custom_format):
    if custom_format:
        custom_format = application.configuration.get_custom_format_string(custom_format)
    application.output.bucket_array(
        application.query_service.find_all_buckets_in_list(list_id), is_json, is_jsonvja, custom_format)


# labels
@cli.group('label', help='subcommand: label (see help)')
def label_group():
    pass


@label_group.command('ls', help='print labels ... (id; title)')
@click.option('is_json', '--json', default=False, is_flag=True,
              help='print as Vikunja json')
@click.option('is_jsonvja', '--jsonvja', default=False, is_flag=True,
              help='print as vja application json')
@click.option('custom_format', '--custom-format',
              help='use formatting string from .vjacli/vja.rc')
@with_application
def label_ls(application, is_json, is_jsonvja, custom_format):
    if custom_format:
        custom_format = application.configuration.get_custom_format_string(custom_format)
    application.output.label_array(
        application.query_service.find_all_labels(), is_json, is_jsonvja, custom_format)


@label_group.command('add', help='add label with title')
@click.argument('title', required=True, nargs=-1)
@with_application
def label_add(application, title):
    label = application.command_service.add_label(" ".join(title))
    click.echo(f'Created label {label.id}')


# tasks
@cli.command('add', help='add new task')
@click.argument('title', required=True, nargs=-1)
@click.option('list_id', '-l', '--folder', '--project', '--list',
              help='list (id or name), default: first favorite list')
@click.option('note', '-n', '--note', '--description',
              help='set description (note)')
@click.option('prio', '-p', '--prio', '--priority',
              help='set priority')
@click.option('due', '-d', '--due', '--duedate', '--due-date', '--due_date',
              help='set due date (supports parsedatetime expressions)')
@click.option('favorite', '-f', '--star', '--favorite', type=click.BOOL,
              help='mark as favorite')
@click.option('tag', '-t', '--tag', '--label',
              help='set label (label must exist on server)')
@click.option('reminder', '-r', '--alarm', '--remind', '--reminder', is_flag=False, flag_value='due',
              help='set reminder (supports parsedatetime expressions). Leave empty to set to due date.')
@click.option('force_create', '--force-create', '--force', is_flag=True,
              help='force creation of non existing label')
@with_application
@click.pass_context
def task_add(ctx, application, title, **args):
    task = application.command_service.add_task(" ".join(title), {k: v for k, v in args.items() if v is not None})
    click.echo(f'Created task {task.id} in list {task.tasklist.id}')
    ctx.invoke(task_show, tasks=[task.id])


@cli.command('edit', aliases=['modify', 'update'],
             help='modify task/tasks (opens task in browser if no options are given)')
@click.argument('task_ids', required=True, type=click.INT, nargs=-1)
@click.option('title', '-i', '--title',
              help='set title')
@click.option('note', '-n', '--note', '--description',
              help='set description (note)')
@click.option('note_append', '-a', '--note-append', '--append-note', '--description-append', '--append-description',
              help='append description to existing note separated by new line')
@click.option('prio', '-p', '--prio', '--priority', type=click.INT,
              help='set priority')
@click.option('list_id', '-l', '--folder-id', '--project-id', '--list-id', '--list_id', type=click.INT,
              help='move to list id')
@click.option('position', '--position', type=click.INT,
              help='set list position')
@click.option('bucket_id', '--bucket-id', '--bucket_id', type=click.INT,
              help='set bucket id')
@click.option('kanban_position', '--kanban-position', '--kanban_position', type=click.INT,
              help='set kanban position')
@click.option('due', '-d', '--due', '--duedate', '--due-date', '--due_date',
              help='set due date (supports parsedatetime expressions)')
@click.option('favorite', '-f', '--favorite', '--star', type=click.BOOL,
              help='mark as favorite')
@click.option('completed', '-c', '--completed', '--done', type=click.BOOL,
              help='mark as completed')
@click.option('tag', '-t', '--tag', '--label',
              help='set label (label must exist on server unless called with --force-create)')
@click.option('reminder', '-r', '--alarm', '--remind', '--reminder', is_flag=False, flag_value='due',
              help='set reminder (supports parsedatetime expressions). Leave empty to set to due date.')
@click.option('force_create', '--force-create', '--force', is_flag=True, default=None,
              help='force creation of non existing label')
@with_application
@click.pass_context
def task_edit(ctx, application, task_ids, **args):
    args_present = {k: v for k, v in args.items() if v is not None}
    for task_id in task_ids:
        if args_present:
            task = application.command_service.edit_task(task_id, args_present)
            click.echo(f'Modified task {task.id} in list {task.tasklist.id}')
            ctx.invoke(task_show, tasks=[task.id])
        else:
            application.open_browser(task_id)


@cli.command('toggle', aliases=['check', 'click', 'done'], help='shortcut for marking / unmarking task as done')
@click.argument('task_id', required=True, type=click.INT)
@with_application
@click.pass_context
def task_toggle(ctx, application, task_id):
    task = application.command_service.toggle_task_done(task_id)
    click.echo(f'Modified task {task.id} in list {task.tasklist.id}')
    ctx.invoke(task_show, tasks=[task_id])


@cli.command('ls', help='list tasks ... (task-id; priority; is_favorite; title; due_date; '
                        'has reminder; namespace; list; labels; urgency)')
@click.option('is_json', '--json', default=False, is_flag=True,
              help='print as Vikunja json')
@click.option('is_jsonvja', '--jsonvja', default=False, is_flag=True,
              help='print as vja application json')
@click.option('custom_format', '--custom-format',
              help='use formatting string from .vjacli/vja.rc')
@click.option('sort_string', '--sort', '--sort-by', '--sort-string',
              help='sort by arguments (task arguments separated by ,) (prefix each with - to reverse order')
@click.option('include_completed', '--include-completed', default=False, is_flag=True,
              help='include completed tasks')
@click.option('bucket_filter', '-b', '--bucket', '--bucket-id', '--bucket_id',
              help='filter by kanban bucket id')
@click.option('due_date_filter', '-d', '--due', '--due-date', '--due_date',
              help='filter by due date. The TEXT value must be like <logical operator> <value>, '
                   'where <logical operator> in eq, gt, lt, ge, le, ne, before, after. '
                   '(supports parsedatetime expressions)')
@click.option('favorite_filter', '-f', '--favorite', '--star', type=click.BOOL,
              help='filter by favorite flag')
@click.option('label_filter', '-t', '--tag', '--label',
              help='filter by label (name or id)')
@click.option('list_filter', '-l', '--list',
              help='filter by list (name or id)')
@click.option('namespace_filter', '-n', '--namespace',
              help='filter by namespace (name or id)')
@click.option('priority_filter', '-p', '--prio', '--priority',
              help='filter by priority. The TEXT value must be like <logical operator> <value>, '
                   'where <logical operator> in eq, gt, lt, ge, le, ne.')
@click.option('title_filter', '-i', '--title',
              help='filter title (regex)')
@click.option('urgency_filter', '-u', '--urgency', is_flag=False, flag_value=3, type=click.INT,
              help='filter by urgency at least')
@with_application
def task_ls(application, is_json, is_jsonvja, custom_format, include_completed, sort_string=None, **filter_args):
    if custom_format:
        custom_format = application.configuration.get_custom_format_string(custom_format)
    filter_args = {k: v for k, v in filter_args.items() if v is not None}

    tasks = application.query_service.find_filtered_tasks(include_completed, sort_string, filter_args)
    application.output.task_array(tasks, is_json, is_jsonvja, custom_format)


@cli.command('show', help='show task details. Multiple task ids may be given')
@click.argument('tasks', type=click.INT, nargs=-1)
@click.option('is_json', '--json', default=False, is_flag=True, help='print as Vikunja json')
@click.option('is_jsonvja', '--jsonvja', default=False, is_flag=True, help='print as vja application json')
@with_application
def task_show(application, tasks, is_json, is_jsonvja):
    for task_id in tasks:
        task = application.query_service.find_task_by_id(task_id)
        application.output.task(task, is_json, is_jsonvja)


@cli.command(name='open', help='open task in browser')
@click.argument('task', required=False, type=click.INT)
@with_application
def task_open(application, task):
    application.open_browser(task)


@cli.command('logout', help='remove local access token')
@with_application
def logout(application):
    application.command_service.logout()


if __name__ == '__main__':
    cli()



import dataclasses
import typing
from dataclasses import dataclass, field
from datetime import datetime, timedelta

import dateutil.parser
from dateutil import tz


def custom_output(cls):
    def str_function(self):
        """Returns a string containing only the non-null attribute values, excluding json attribute ."""
        return '\n'.join(f'{attribute.name}: {_str_value(getattr(self, attribute.name))}'
                         for attribute in dataclasses.fields(self)
                         if attribute.name != 'json' and getattr(self, attribute.name))

    def _str_value(attribute_value):
        return [_str_value(x) for x in attribute_value] if isinstance(attribute_value, list) else str(attribute_value)

    setattr(cls, '__str__', str_function)
    return cls


def data_dict(cls):
    def data_dict_function(self):
        return {k: _transform_value(v) for k, v in self.__dict__.items() if k != 'json'}

    def _transform_value(v):
        if _is_data_dict(v):
            return v.data_dict()
        if isinstance(v, list):
            return [_transform_value(x) for x in v]
        return v

    def _is_data_dict(v):
        return hasattr(v, 'data_dict') and callable(v.data_dict)

    setattr(cls, 'data_dict', data_dict_function)
    return cls


@dataclass(frozen=True)
@data_dict
class User:
    json: dict = field(repr=False)
    id: int
    username: str
    name: str
    default_list_id: int

    @classmethod
    def from_json(cls, json):
        return cls(json, json['id'], json['username'], json['name'], json['settings']['default_list_id'])


@dataclass(frozen=True)
@data_dict
class Namespace:
    json: dict = field(repr=False)
    id: int
    title: str
    description: str
    is_archived: bool

    @classmethod
    def from_json(cls, json):
        return cls(json, json['id'], json['title'], json['description'], json['is_archived'])

    @classmethod
    def from_json_array(cls, json_array):
        return [Namespace.from_json(x) for x in json_array or []]


@dataclass(frozen=True)
@data_dict
class List:
    json: dict = field(repr=False)
    id: int
    title: str
    description: str
    is_favorite: bool
    is_archived: bool
    namespace: Namespace

    @classmethod
    def from_json(cls, json, namespace):
        return cls(json, json['id'], json['title'], json['description'], json['is_archived'], json['is_favorite'],
                   namespace)

    @classmethod
    def from_json_array(cls, json_array, namespace):
        return [List.from_json(x, namespace) for x in json_array or []]


@dataclass(frozen=True)
@data_dict
class Bucket:
    json: dict = field(repr=False)
    id: int
    title: str
    is_done_bucket: bool
    limit: int
    position: int
    count_tasks: int

    @classmethod
    def from_json(cls, json):
        return cls(json, json['id'], json['title'],
                   json['is_done_bucket'],
                   json['limit'],
                   json['position'],
                   len(json['tasks']) if json['tasks'] else 0)

    @classmethod
    def from_json_array(cls, json_array):
        return [Bucket.from_json(x) for x in json_array or []]


@dataclass(frozen=True)
@data_dict
class Label:
    json: dict = field(repr=False)
    id: int
    title: str

    @classmethod
    def from_json(cls, json):
        return cls(json, json['id'], json['title'])

    @classmethod
    def from_json_array(cls, json_array):
        return [Label.from_json(x) for x in json_array or []]


@dataclass
@custom_output
@data_dict
# pylint: disable=too-many-instance-attributes
class Task:
    json: dict = field(repr=False)
    id: int
    title: str
    description: str
    priority: int
    is_favorite: bool
    due_date: datetime
    reminder_dates: typing.List[datetime]
    repeat_mode: int
    repeat_after: timedelta
    start_date: datetime
    end_date: datetime
    percent_done: float
    done: bool
    done_at: datetime
    labels: typing.List[Label]
    tasklist: List
    position: int
    bucket_id: int
    kanban_position: int
    created: datetime
    updated: datetime
    urgency: float = field(init=False)

    def __post_init__(self):
        self.urgency = Urgency.compute(self)

    @property
    def label_titles(self):
        return ",".join(map(lambda label: label.title, self.labels or []))

    @property
    def sortable_due_date(self):
        return int((self.due_date or datetime.max).strftime("%Y%m%d%H%M%S"))

    @classmethod
    def from_json(cls, json, list_object, labels):
        return cls(json, json['id'], json['title'], json['description'],
                   json['priority'],
                   json['is_favorite'],
                   _date_from_json(json['due_date']),
                   [_date_from_json(reminder) for reminder in json['reminder_dates'] or []],
                   json['repeat_mode'],
                   timedelta(seconds=json['repeat_after']),
                   _date_from_json(json['start_date']),
                   _date_from_json(json['end_date']),
                   json['percent_done'],
                   json['done'],
                   _date_from_json(json['done_at']),
                   labels,
                   list_object,
                   json['position'],
                   json['bucket_id'],
                   json['kanban_position'],
                   _date_from_json(json['created']),
                   _date_from_json(json['updated']),
                   )

    def has_label(self, label):
        return any(x.id == label.id for x in self.labels)


urgency_score_coefficients = {'due_date': 1.0, 'priority': 1.0, 'favorite': 1.0, 'list_keyword': 1.0,
                              'label_keyword': 1.0}


class Urgency:

    @staticmethod
    def compute(task: Task):
        if task.done:
            return 0
        due_date_score = Urgency.get_due_date_score(task) * urgency_score_coefficients['due_date']
        priority_score = task.priority * urgency_score_coefficients['priority']
        favorite_score = int(task.is_favorite) * urgency_score_coefficients['favorite']
        list_name_score = int('next' in task.tasklist.title.lower()) * urgency_score_coefficients['list_keyword']
        lable_name_score = int(any('next' in label.title.lower()
                                   for label in task.labels)) * urgency_score_coefficients['label_keyword']

        return 1 + due_date_score + priority_score + favorite_score + list_name_score + lable_name_score

    @staticmethod
    def get_due_date_score(task):
        if task.due_date:
            due_days = (task.due_date - datetime.today()).days
            if due_days < 0:
                result = 6
            elif due_days == 0:
                result = 5
            elif due_days == 1:
                result = 4
            elif 1 < due_days <= 2:
                result = 3
            elif 2 < due_days <= 5:
                result = 2
            elif 5 < due_days <= 10:
                result = 1
            elif due_days > 14:
                result = -1
            else:
                result = 0
        else:
            result = 0
        return result


def _date_from_json(json_date):
    if json_date and json_date > '0001-01-01T00:00:00Z':
        return dateutil.parser.isoparse(json_date).astimezone(tz.tzlocal()).replace(tzinfo=None)
    return None



import logging
import operator
import re

from vja import parse

logger = logging.getLogger(__name__)


def _create_bucket_filter(value):
    logger.debug("filter bucket_id %s", value)
    return lambda x: x.bucket_id == int(value)


def _create_due_date_filter(value: str):
    if value.strip() == '':
        return lambda x: not x.due_date
    arguments = value.split(" ", 1)
    operator_name = arguments[0]
    operation = _operators[operator_name]
    value = parse.parse_date_arg_to_datetime(arguments[1])
    logger.debug("filter due date %s %s", operation.__name__, value)
    return lambda x: (operation(x.due_date, value) if x.due_date else False)


def _create_favorite_filter(value):
    logger.debug("filter favorite %s", value)
    return lambda x: x.is_favorite == bool(value)


def _create_label_filter(value):
    logger.debug("filter labels %s", value)
    if str(value).isdigit():
        return lambda x: any(label.id == int(value) for label in x.labels)
    if str(value).strip() == '':
        return lambda x: not x.labels
    return lambda x: any(label.title == value for label in x.labels)


def _create_list_filter(value):
    logger.debug("filter list %s", value)
    if str(value).isdigit():
        return lambda x: x.tasklist.id == int(value)
    return lambda x: x.tasklist.title == value


def _create_namespace_filter(value):
    logger.debug("filter namespace %s", value)
    if str(value).isdigit():
        return lambda x: x.tasklist.namespace.id == int(value)
    return lambda x: x.tasklist.namespace.title == value


def _create_title_filter(value):
    return lambda x: bool(re.search(re.compile(value), x.title))


def _create_priority_filter(value):
    arguments = str(value).split(" ", 1)
    operator_name = arguments[0]
    operation = _operators[operator_name]
    value = arguments[1]
    logger.debug("filter priority %s %s", operation.__name__, value)
    return lambda x: operation(x.priority, int(value))


def _create_urgency_filter(value):
    logger.debug("filter urgency %s", value)
    return lambda x: x.urgency >= value


_operators = {
    'eq': operator.eq,
    'ne': operator.ne,
    'lt': operator.lt,
    'le': operator.le,
    'gt': operator.gt,
    'ge': operator.ge,
    'before': operator.lt,
    'after': operator.gt
}

_filter_mapping = {
    'bucket_filter': _create_bucket_filter,
    'due_date_filter': _create_due_date_filter,
    'favorite_filter': _create_favorite_filter,
    'label_filter': _create_label_filter,
    'list_filter': _create_list_filter,
    'namespace_filter': _create_namespace_filter,
    'title_filter': _create_title_filter,
    'priority_filter': _create_priority_filter,
    'urgency_filter': _create_urgency_filter,
}


def create_filters(filter_args):
    filters = []
    for filter_name, filter_value in filter_args.items():
        filters.append(_filter_mapping[filter_name](filter_value))
    return filters
